#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sstream>
#include <vector>
#include <deque>
#include <map>

#include <sys/wait.h>
#include <unistd.h>


#include "types.h"
extern "C" {
#include "state.h"
#include "movie.h"
}
#include "fceu.h"

#define BISQBOT_INTERNAL
#include "bisqbotdefs.hh"
#include "bisqbot.hh"

#include "coroutine.h"

int BotFrontDisableVideo=0;

#define INCLUDE_BOT_IN_SAVESTATE 0
struct SaveState
{
    std::vector<unsigned char> Data;

#if INCLUDE_BOT_IN_SAVESTATEA
    const BisqBotStateBase* Bot;
    SaveState() : Bot(0) { }
    ~SaveState() { delete Bot; }
#endif

    void Create()
    {
#if INCLUDE_BOT_IN_SAVESTATEA
        delete Bot;
#endif
        
        FILE*fp = tmpfile();
        FCEUSS_SaveFP(fp);
        rewind(fp);
        
        fseek(fp,0,SEEK_END);
        long pos = ftell(fp);
        rewind(fp);
        
        SaveState result;
        Data.resize(pos);
        fread(&Data[0], 1, pos, fp);
        fclose(fp);
        
#if INCLUDE_BOT_IN_SAVESTATEA
        Bot = BisqBotNewSaveState();
#endif
    }
    void Load()
    {
        FILE*fp = tmpfile();
        fwrite(&Data[0], 1, Data.size(), fp);
        rewind(fp);
        FCEUSS_LoadFP(fp, 0);
        fclose(fp);
#if INCLUDE_BOT_IN_SAVESTATEA
        BisqBotLoadState(Bot);
#endif
    }
#if INCLUDE_BOT_IN_SAVESTATEA
private:
    SaveState(const SaveState& b);
    void operator=(const SaveState& b);
#endif
};

static unsigned char CurInput;
static bool Active=false;

static void DisplaySeq(const std::vector<unsigned char>& Seq)
{
    std::stringstream msg;
    for(unsigned a=0; a<Seq.size(); )
    {
        unsigned c=1, ch=Seq[a];
        while(a+c < Seq.size() && Seq[a+c] == ch) ++c;
        msg << c; msg << '*';
        
        unsigned b=0;
        if(ch & K_A) { if(b++) msg << ("_"); msg << ("A"); }
        if(ch & K_B) { if(b++) msg << ("_"); msg << ("B"); }
        if(ch & K_SE) { if(b++) msg << ("_"); msg << ("SE"); }
        if(ch & K_ST) { if(b++) msg << ("_"); msg << ("ST"); }
        if(ch & K_U) { if(b++) msg << ("_"); msg << ("U"); }
        if(ch & K_D) { if(b++) msg << ("_"); msg << ("D"); }
        if(ch & K_L) { if(b++) msg << ("_"); msg << ("L"); }
        if(ch & K_R) { if(b++) msg << ("_"); msg << ("R"); }
        if(!b) msg << ("00");
        msg << (", ");
        a += c;
    }
    BotFrontMsg("%s", msg.str().c_str());
}

static std::vector<unsigned char> GlobalSeqSample;
void BisqBotDisplaySeqSample()
{
    DisplaySeq(GlobalSeqSample);
    BotFrontMsg("\n");
}


#include <sys/file.h> /* For file locking */
void BotFrontMsg(const char*fmt, ...)
{
    FILE*fp;
    
    va_list ap;
    va_start(ap,fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    
    fp = fopen("rockman.log", "at");
    if(!fp) fp = fopen("rockman.log", "wt");
    if(!fp) return;
    flock(fileno(fp), LOCK_EX);
    fseek(fp,0,SEEK_END);

    va_start(ap,fmt);
    vfprintf(fp, fmt, ap);
    va_end(ap);
    
    fclose(fp);     
}

void BotFrontStart()
{
    fprintf(stderr,"BotFrontStart\n");
    BotFrontDisableVideo=1;
    Active=true;
}

void BotFrontEnd()
{
    fprintf(stderr,"BotFrontEnd\n");
    BisqBotReset();
    BotFrontDisableVideo=0;
    Active=false;
} 

void BotFrontToggle()
{
    if(!Active)
        BotFrontStart();
    else
        BotFrontEnd();
}

int BotFrontActive()
{
    return Active ? 1 : 0;
}

unsigned BotFrontGetKeymask()
{
    if(Active)
    {
        //fprintf(stderr, "CurInput: %02X\n", CurInput);
        return CurInput;
    }
    //fprintf(stderr, "Nul keymask\n");
    return 0;
}

/*
    FCEU main loop:
        1. Emulate frame     
        2. Read keyboard input (BotFrontIter)
        3. Update joypad input (BotFrontGetKeymask)

    It is actually possible for keyboard input to be
    read multiple times per frame... Currently we hope
    this don't happen.
*/

int BqtBossHit=0;
int BqtSelPushed=0;
int BqtBossRemHp=0;
int BqtRockY, BqtRockXS, BqtFiring;
int BqtDamaged, BqtMegaKilled;
int BqtScrolled;

extern uint32_t framecount;

#include "averages.hh"
#include "mutation.hh"
#include "hash.hh"

struct CutmanGlitchNavigatorv2
{
    // How many frames to analyze each time
    static const unsigned Try_length   = 25;
    // How many frames to actually preserve
    static const unsigned Keep_length  = 6;
    
    // Define the range of parameters to be passed to Mutate().
    static const double MINMUTAATIO = 0.0010; // Minimum value, given to first candidates
    static const double MAXMUTAATIO = 6.4000; // Maximum value, given to last candidates
    // This controls the curve of mutation distribution 
    // (larger value = more big mutations, smaller value = more small mutations).
    // 1 = even distribution
    static const double LOGMUTAATIO = 0.4;
    
    static const unsigned WINNERS = 16; // number of winners to take each round for refinement

    // number of past-round winners to take each round, in case they were good but unlucky
    // (should only be used when the tasks given to the candidates are not identical)
    static const unsigned OLD_WINNERS = 0;

    // number of variations per each generation (PLUS WINNERS, which are preserved for competition)
    // larger number increases the chances of finding the Right Thing
    static const unsigned NUM_VARIATIONS = 900;
    
    // number of entirely randomly generated candidates to add each round
    static const unsigned NUM_RANDOMS = 400;

    // accept rule when no improvements have been found within this number of generations
    // larger number decreases the chances of missing the Right Thing
    static const unsigned min_generations_no_improvement = 80;

    static const unsigned NUM_FORKS = 2;
    static const unsigned NUM_CANDIDATES_PER_PROCESS = 300;

    template<unsigned Length>
    struct RandomInputCandidate
    {
        struct InputType
        {
            float A, B, Dir;
            //InputType() : A(0),B(0),Dir(0) { }
            
            static unsigned CountStates() { return 4*3; }
            void SetState(unsigned n)
            {
                A = (n%2 + 0.5) / 2.0; n/=2;
                B = (n%2 + 0.5) / 2.0; n/=2;
                Dir = (n%3 + 0.5) / 3.0; n/=3;
            }
            
            static unsigned CountSimpleMutations() { return 4; }
            void DoSimpleMutation(unsigned n)
            {
                switch(n)
                {
                    case 0: A=fmod(A + 1/2.0,1); break; // Flip A
                    case 1: B=fmod(B + 1/2.0,1); break; // Flip B
                    case 2: Dir = fmod(Dir+ 1/3.0, 1); break; // Change Dir to other value
                    case 3: Dir = fmod(Dir+ 2/3.0, 1); break; // Change Dir to another value
                }
            }
            
            const unsigned char GetKey() const
            {
                unsigned char res = 0;
                switch(((unsigned)(A*2))%2) { case 1: res |= K_A; }
                switch(((unsigned)(B*2))%2) { case 1: res |= K_B; }
                switch(((unsigned)(Dir*3))%3) { case 1: res |= K_R; break;
                                                case 2: res |= K_L; break; }
                return res;
            }
            
            void Mutate(double mut_amount)
            {
                unsigned r = lrand48();
                if(r & 0x0000F00F) // Chance: 10 bits
                    A = fmod(A+mut_amount, 1);
                if(r & 0x003F03F0) // Chance: 12 bits
                    Dir = fmod(Dir+mut_amount, 1);
                if(r & 0x04000400) // Chance: 2 bits
                    B = fmod(B+mut_amount, 1);
            }
            
            void Dump() const
            {
                printf("{%g,%g,%g}", A,B,Dir);
            }
        };

        InputType input[Length];
        
        double scoring,RockX; int scrolled, keepable;
        unsigned confidence;
        unsigned key;

        void Dump() const
        {
            printf("{ {");
            for(unsigned a=0; a<Length; ++a)
            {
                printf("%s", a?",":"");
                input[a].Dump();
            }
            printf("}, 0}");
        }
        
        static unsigned CountSimpleMutations()
        {
            return InputType::CountSimpleMutations() * Length;
        }
        
        void DoSimpleMutation(unsigned n)
        {
            input[n % Length].DoSimpleMutation(n / Length);
        }
        
        void Mutate(double amount)
        {
            amount *= Length;
            while(amount > 0.0001)
            {
                unsigned ind = lrand48() % Length;
                unsigned method = 0;
                if(amount > 0.2) method = lrand48() % 4;
                double mut_amount =
                    std::min(amount, std::min(0.95, 0.03+drand48()));
                /* Maximum of above is 0.95. */
                switch(method)
                {
                    case 0:
                    case 1:
                    { DoMut:
                        input[ind].Mutate(mut_amount);
                        break;
                    }
                    case 2: // insert a random item
                    {
                        for(unsigned b=Length-1; b > ind; --b)
                            input[b] = input[b-1];
                        goto DoMut;
                    }
                    case 3: // delete a random item
                    {
                        for(unsigned b=ind; b+1<Length; ++b)
                            input[b] = input[b+1];
                        ind = Length-1;
                        goto DoMut;
                    }
                }
                amount -= mut_amount;
            }
        }
        
        // transform the floating point value (0..<1) into an actual input
        const unsigned char get(unsigned ind) const
        {
            return input[ind].GetKey();
        }
        
        void DeleteFirst(unsigned n)
        {
            if(n > Length) return;
            unsigned remain = Length-n;
            for(unsigned a=0; a<remain; ++a)
                input[a] = input[a+n];
        }
        RandomInputCandidate<Length> CreateSeed(unsigned type)
        {
            scoring=0; confidence=0; key=0;
            for(unsigned a=0; a<Length; ++a)
                input[a].SetState(type);
            return *this;
        }
        
        RandomInputCandidate<Length> CreateRandom()
        {
            unsigned d = lrand48() % InputType::CountStates();
            for(unsigned pos=0; pos < Length; ++pos)
            {
                input[pos].SetState(d);
                if(!(lrand48()%5)) d = lrand48() % InputType::CountStates();
            }
            return *this;
        }
        
        bool IsSameAs(const RandomInputCandidate<Length>& b) const
        {
            for(unsigned a=0; a<Length; ++a)
                if(get(a) != b.get(a)) return false;
            return true;
        }
        bool IsSimplerThan(const RandomInputCandidate<Length>& b) const
        {
            unsigned complx_me=0, complx_b=0;
            for(unsigned a=0; a<Length; ++a)
            {
                complx_me += CountInputComplexity(get(a));
                complx_b  += CountInputComplexity(b.get(a)); 
            }
            return complx_me < complx_b;
        }
        
        static unsigned CountInputComplexity(unsigned char c)
        {
            unsigned res=0;
            for(unsigned a=1; a<=128; a*=2) if(c&a) ++res;
            return res;
        }
        
        /* Although this is "operator<", it really returns
         * whether this is "better" than the other, i.e.
         * higher score etc.
         */
        bool operator< (const RandomInputCandidate<Length>& b) const
        {
            if(scoring != b.scoring) return scoring > b.scoring;
            return IsSimplerThan(b);
        }
        
        const std::string GetAsString() const
        {
            unsigned char Buf[Length];
            for(unsigned a=0; a<Length; ++a)
                Buf[a] = get(a);
            return std::string(Buf, Buf+Length);
        }
    };

    void Run()
    {
        #define RunFrameWith(k)  \
            do { \
                CurInput = (k); FCEUI_FrameAdvance(); scrReturnV; \
                if(!Active)goto EndLoop; \
            } while(0)

        scrBegin;

        typedef RandomInputCandidate<Try_length> CandidateType;
        typedef std::deque<CandidateType> CandidateListType;
        static CandidateListType WinnerList;
        
        WinnerList.clear();

#if 0
        static const CandidateType model =
{ {{0,0,0.3},
   {0,0,0.3},
   {0,0,0.7},
   {0,0,0.3},
   {0,0,0.3},

   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},

   {0,0,0.7},
   {0,0,0.3},
   {0,0,0.7},
   {0,0,0.7},
   {0,0,0.3},

   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3},
   {0,0,0.3}
   }}
       ;WinnerList.push_back(model);
#endif
        
        /* Every iteration (Keep_length frames long) */
        for(;;)
        {
            /* Append the seed group to this list */
            for(unsigned a=0; a< CandidateType::InputType::CountStates(); ++a)
                WinnerList.push_back(CandidateType().CreateSeed(a));
            
            static SaveState itbegin;
            static double begin_x;
            static double best_scoring;
            static CandidateType best_candidate;
            static SaveState best_progress_snapshot;
            itbegin.Create();

            static unsigned generationno;
            static unsigned n_generations_no_improvement;

            begin_x = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
            best_scoring = -99999;
            n_generations_no_improvement = 0;

            static hash_set<std::string> covered_candidates;
            covered_candidates.clear();
            
            /* Generate NUM_VARIATIONS for as many generations as needed */
            for(generationno=1; ; ++generationno)
            {
                static CandidateListType CandidateList;
                /* generate new scorings from models */
                CandidateList = WinnerList;
                
                for(unsigned b=0; b<WinnerList.size(); ++b)
                for(unsigned a=0; a<CandidateType::CountSimpleMutations(); ++a)
                {
                    CandidateType tmp = WinnerList[b];
                    tmp.DoSimpleMutation(a);
                    tmp.confidence = 0;
                    
                    std::string c = tmp.GetAsString();
                    if(covered_candidates.find(c) != covered_candidates.end()) continue;
                    covered_candidates.insert(c);
                    
                    CandidateList.push_back(tmp);
                }
                
                if(n_generations_no_improvement > 0 || generationno==1)
                for(unsigned a=0; a<NUM_VARIATIONS; ++a)
                {
                    // Take one of the best as a model, and mutate a copy of it
                    unsigned randi = lrand48() % WinnerList.size();
                    CandidateType tmp = WinnerList[randi];

                    static const double minlog  = pow(LOGMUTAATIO, MINMUTAATIO);
                    static const double maxlog  = pow(LOGMUTAATIO, MAXMUTAATIO);
                    const double logpos = minlog + (a*(maxlog-minlog))/(NUM_VARIATIONS-1);
                    const double mut_amount = log(logpos) / log(LOGMUTAATIO);
                    
                    //if(generationno == 1)
                    //    fprintf(stderr, "Gen %u: mut_amount=%g\n", a, mut_amount);

                    tmp.confidence = 0;
                    tmp.Mutate(mut_amount);

                    std::string c = tmp.GetAsString();
                    if(covered_candidates.find(c) != covered_candidates.end()) continue;
                    covered_candidates.insert(c);

                    CandidateList.push_back(tmp);
                }
                //if(generationno==1){Active=0;}
                
                if(n_generations_no_improvement > 0 || generationno==1)
                for(unsigned a=0; a<NUM_RANDOMS; ++a)
                {
                    CandidateType tmp; tmp.CreateRandom();

                    std::string c = tmp.GetAsString();
                    if(covered_candidates.find(c) != covered_candidates.end()) continue;
                    covered_candidates.insert(c);

                    CandidateList.push_back(tmp);
                }

                WinnerList.clear(); // list of old winners not needed at this point
                
                ++n_generations_no_improvement;
                    
                static std::map<int, unsigned> pid_list;
                pid_list.clear();
                static unsigned candno;
                BotFrontDisableVideo=2;
                for(candno=0; candno<CandidateList.size(); candno+=NUM_CANDIDATES_PER_PROCESS)
                {
                    static int pid;
                    {int waitflag = (pid_list.size() < NUM_FORKS) ? WNOHANG : 0;
                    do {
                        int status=0;
                        int pid = waitpid(-1, &status, waitflag);
                        if(pid > 0)
                        {
                            if(WIFSIGNALED(status))
                            {
                                fprintf(stderr, "Child died at signal %d\n",
                                    WTERMSIG(status));
                            }
                            pid_list.erase(pid);
                        }
                        waitflag=0;
                    } while(pid_list.size() >= NUM_FORKS); }
                    
                    fflush(stdout);
                    fflush(stderr);
                    pid = fork();
                    if(pid > 0)
                    {
                        pid_list[pid] = candno;
                        fprintf(stderr, "."); fflush(stderr);
                        continue;
                    }
                    if(pid < 0)
                    {
                        perror("fork");
                    }
                    MovieForkIntoTempFiles();
                    
                    static unsigned candcounter;
                    for(candcounter=0; candno<CandidateList.size()
                                    && candcounter<NUM_CANDIDATES_PER_PROCESS;
                                    ++candcounter, ++candno)
                    {
                        static CandidateType* candit;
                        candit = &CandidateList[candno];

                        itbegin.Load();
                        /*
                        fprintf(stderr, "Trying(%u) input:", (unsigned)framecount);
                        for(frameno=0; frameno<Try_length; ++frameno)
                            fprintf(stderr, " %02X", candit->get(frameno));
                        */
                        
                        candit->scoring = 0.0;
                        candit->scrolled= 0;
                        BqtScrolled=0;
                        BqtMegaKilled=0;
                        static unsigned frameno;
                        for(frameno=0; frameno<Try_length; ++frameno)
                        {
                            //fprintf(stderr, "pid %d: frame %u\n", getpid(), frameno);
                            BqtMegaKilled=0;
                            RunFrameWith(candit->get(frameno));
                            if(BqtMegaKilled) goto BadScore; // death occurred!
                            if(BqtScrolled)
                            {
                                // ycoordinate hack
                                //if(RAM[0x460] >= 0x29) goto BadScore;
                                if(frameno+1 < Try_length)
                                    ++frameno; // break the NEXT frame to prevent Infinity
                                break;
                            }
                        }
                        static unsigned keepable_frames;
                        keepable_frames = frameno < Keep_length ? frameno : Keep_length;
                        
                        /* TODO: In the vertical fall in Wily2, calculate score
                         * by the horizontal distance from the spot where the
                         * teleporter appear.
                         */
                         
                        /* Clone robot room:
                         *     Optimal X = 10576
                         */
                        
                        static double RockX;
                        RockX = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
                        
                        //if(!BqtScrolled && RockX >= 8312) goto BadScore; // TOO FAR
                        
                        if(BqtScrolled)
                            RockX /*+*/= 256 + 5*(Try_length-frameno); // Award scrolling justly!

                        if(BqtScrolled)
                        {
                            // Minimal distance to Clone robot's spot is important
                            if(RAM[0x460] < 0x28)
                                RockX = RockX*16 - std::abs(RAM[0x480] + RAM[0x4A0]/256.0
                                                  - 0x36)*16;
                            else
                            {
                                unsigned sweetspot = 0x50;
                                if(RAM[0x600] < 252) RockX=1;
                                else RockX -= (sweetspot - RAM[0x480]) & 255;
                                if(!(RAM[0x420] & 0x40)) goto BadScore; // should be facing right
                            }
                         //   candit->scoring = RockX;
                        }
                        else
                            candit->scoring = 0;//(RockX - begin_x) / frameno;
                        //if(BqtScrolled) candit->scoring = RAM[0x600]; - ycoordinate hack
                        // Extra award for scrolling on how soon it happened!
                        //if(BqtScrolled) candit->scoring += (Try_length-frameno)*1.1;
                        candit->confidence++;
                    BadScore:
                        char candfn[128]; sprintf(candfn, "/tmp/cand%d.res", (int)candno);
                        FILE*fp = fopen(candfn, "wt");
                        if(fp)
                        {
                            fprintf(fp, "%g,%d,%g,%d,%d\n",
                                candit->scoring,
                                candit->confidence,
                                RockX,BqtScrolled,keepable_frames);
                            fclose(fp);
                        }
                        else
                            perror(candfn);
                    }
                    fflush(stdout);
                    fflush(stderr);
                    _exit(0);
                }
                BotFrontDisableVideo=1;

                while(pid_list.size() > 0)
                {
                    int status=0;
                    int pid = waitpid(-1, &status, 0);
                    if(pid > 0)
                    {
                        if(WIFSIGNALED(status))
                        {
                            fprintf(stderr, "Child died at signal %d\n",
                                WTERMSIG(status));
                        }
                        pid_list.erase(pid);
                    }
                }
                
                fprintf(stderr, "\r");
                /* Read the results from the tasks */
                for(candno=0; candno<CandidateList.size(); ++candno)
                {
                    CandidateType* candit = &CandidateList[candno];
                    candit->key = candno;

                    char candfn[128]; sprintf(candfn, "/tmp/cand%d.res", (int)(candno));
                    if(FILE*fp = fopen(candfn, "rt"))
                    {
                        fscanf(fp, "%lg,%d,%lg,%d,%d",
                            &candit->scoring,
                            &candit->confidence,
                            &candit->RockX,
                            &candit->scrolled,
                            &candit->keepable);
                        fclose(fp);
                    }
                    else
                    {
                        perror(candfn);
                    }
                    unlink(candfn);
                }

                // Sort the new-round candidates, best-first
                std::sort(CandidateList.begin(), CandidateList.end());
                
                /* Check if there was a new winner */
                for(candno=0; candno<CandidateList.size(); ++candno)
                {
                    static CandidateType* candit;
                    candit = &CandidateList[candno];

                    if(candit->scoring > best_scoring)
                    {
                    MakeNewBest:
                        fprintf(stderr, "New record [%u/%u]: %g (RockX=%g%s)\n",
                            candit->key, (unsigned)CandidateList.size(),
                            candit->scoring,
                            candit->RockX, candit->scrolled?" SCROLLED":"");

                        {std::vector<unsigned char> seq;
                        for(unsigned frameno=0; frameno < Try_length; ++frameno)
                            seq.push_back(candit->get(frameno));
                        DisplaySeq(seq);}
                        printf("\n");
                        candit->Dump();
                        printf("\n");
                    
                        best_scoring   = candit->scoring;
                        best_candidate = *candit;
                        
                        // Replay the first N frames of the input and make a snapshot
                        itbegin.Load();
                        static unsigned frameno;
                        for(frameno=0; frameno < candit->keepable; ++frameno)
                            RunFrameWith(candit->get(frameno));
                        
                        best_progress_snapshot.Create();
                        // save a progress snapshot (savestate) on disk for manual inspection
                        FCEUSS_Save("botfront2-middle");
                        
                        // reset the no-progress counter
                        n_generations_no_improvement = 0;
                    }
                    else if(candit->scoring == best_scoring)
                    {
                        // If this candidate has the same score as the winner,
                        // check if this one is simpler.
                        
                        if(candit->IsSimplerThan(best_candidate))
                            goto MakeNewBest;
                        else if(best_candidate.IsSimplerThan(*candit))
                            candit->scoring /= 2.0; // Forget it!
                    }
                }
                
                std::sort(CandidateList.begin(), CandidateList.end());

                // Regenerate the inheritor list.
                
                {unsigned keycode=0;
                for(CandidateListType::iterator
                      i=CandidateList.begin();  
                      i!=CandidateList.end(); ++i)
                {
                    i->key = keycode++;
                }}

                if(OLD_WINNERS > 0)
                {
                    // Pass 1: Take the previous-pass winners that still are living.
                    for(CandidateListType::iterator
                          i=CandidateList.begin();  
                          i!=CandidateList.end(); ++i)
                    {
                        if(i->confidence > 1) // >1 signifies a previous-pass winner
                        {
                            // It loses 1 life, but still has a chance
                            CandidateType tmp = *i;
                            --tmp.confidence;
                            WinnerList.push_back(tmp);
                        }
                    }
                    // Sort them, best first.
                    std::sort(WinnerList.begin(), WinnerList.end());

                    // Crop the previous-pass winners list to the maximum size.
                    if(WinnerList.size() > OLD_WINNERS)
                        WinnerList.erase(WinnerList.begin() + OLD_WINNERS,
                                           WinnerList.end());
                }
                
                if(WINNERS > 0)
                {
                    unsigned take_winners = WINNERS;
                    if(!n_generations_no_improvement) take_winners = 1;
                    
                    // Pass 2: Take best candidates from the new round
                    unsigned newwinners = 0;
                    for(CandidateListType::const_iterator i=CandidateList.begin();
                          newwinners < take_winners && i!=CandidateList.end(); ++i)
                    {
                        CandidateListType::const_iterator j;
                        
                        /* Ensure that the candidate in question wasn't already
                         * included in the list (previous-round winners)
                         */
                        for(j=WinnerList.begin(); j!=WinnerList.end(); ++j)
                        {
                            if(j->key == i->key) goto Already_Had;
                            if(i->IsSameAs(*j)) goto Already_Had;
                        }
                        
                        WinnerList.push_front(*i);
                        ++newwinners;
                    Already_Had: ;
                    }
                }
                 
                Averaging quality_all, quality_best;
                for(CandidateListType::const_iterator i=WinnerList.begin();
                      i!=WinnerList.end(); ++i)
                    quality_best.Cumulate(i->scoring);
                for(CandidateListType::const_iterator i=CandidateList.begin();
                      i!=CandidateList.end(); ++i)
                    quality_all.Cumulate(i->scoring);
                
                best_progress_snapshot.Load();
                fprintf(stderr, "Average quality of generation %u @ %u: %g (all), %g (inheritors)\n",
                    generationno, (unsigned)framecount,
                    quality_all.GetValue(), quality_best.GetValue());
    
                if(n_generations_no_improvement >= min_generations_no_improvement) break;
                
                /*if(n_generations_no_improvement >= 6
                && quality_best.GetValue() == 0.0)
                {
                    fprintf(stderr, "Giving up, nothing seems to happen in this frame\n");
                    break;
                }*/
            } // proceed to next generation
            
            covered_candidates.clear();
            
            fprintf(stderr, "- next %u frames!\n", Keep_length);
            best_progress_snapshot.Load();
            
            for(CandidateListType::iterator i=WinnerList.begin();
                  i!=WinnerList.end(); ++i)
                i->DeleteFirst(Keep_length);
        }
    EndLoop: ;

        scrFinishV;
        #undef RunFrameWith
    }
};

struct CutmanGlitchNavigatorv1
{
    void Run()
    {
        scrBegin;
        
        static const unsigned Try_length   = 20;
        static const unsigned Keep_length  = 20;
        static const unsigned Try_maxcount = 12000;
        
        #define RunFrameWith(k)  \
            do { \
                CurInput = (k); FCEUI_FrameAdvance(); scrReturnV; \
                if(!Active)goto EndLoop; \
            } while(0)

        while(Active)
        {
            static std::vector<unsigned char> best_input, cur_input;
            static unsigned trycount;
            static unsigned trypos;
            static SaveState middle_state;
            static double best_x, begin_x;
            static SaveState best_middle;
            static SaveState phasebegin;
            
            best_input.clear(); cur_input.clear();
            
            cur_input.resize(Try_length, 0xff);
            best_input.resize(Try_length, 0xff);
            
            phasebegin.Create();
            
            begin_x  = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
            
            best_x = 0;
        Encouraged:
            for(trycount=0; trycount < Try_maxcount; ++trycount)
            {
                fprintf(stderr, "\rtry %u/%u, best_x=%-10g (avg. speed %-10g)",
                    trycount, Try_maxcount, best_x,
                        (best_x - begin_x) / (double)Try_length);
        RestartTry:
                phasebegin.Load();

                BqtScrolled=0;
                for(trypos=0; trypos<Try_length; ++trypos)
                {
                    if(trypos == Keep_length) middle_state.Create();

                    BqtMegaKilled=0;
                    cur_input[trypos] = BisqBotInput(trypos);
                    RunFrameWith(cur_input[trypos]);

                    if(BqtMegaKilled) goto RestartTry; // death occurred!
                    //double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
                    //if(RockY >= 248.0) goto RestartTry; // death occurred!
                }
                double RockX  = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
                if(BqtScrolled) RockX += 256;
                if(RockX > best_x)
                {
                    best_x = RockX;
                    if(std::memcmp(&best_input[0], &cur_input[0], Keep_length) == 0)
                    {
                        best_input = cur_input;
                        fprintf(stderr, "Same stub though.\n");

             //       if( (double)(best_x - begin_x) / (double)Try_length
               //        >= 1.3124) goto DoneThisStep;
                    }
                    else
                    {
                        best_input = cur_input;
                        best_middle = middle_state;
                        best_middle.Load();
                        FCEUSS_Save("botfront-middle");

             //       if( (double)(best_x - begin_x) / (double)Try_length
               //        >= 1.3124) goto DoneThisStep;

                        goto Encouraged;
                    }
                }
            }
            
            DoneThisStep:
            best_middle.Load();
            fprintf(stderr, " - next!\n");
        }
    EndLoop: ;
        scrFinishV;
        #undef RunFrameWith
   }
};

struct WilyBossFighter
{
    void Run()
    {
        #define RunFrameWith(k)  \
            do { \
                CurInput = (k); FCEUI_FrameAdvance(); scrReturnV; \
                if(!Active)goto EndLoop; \
            } while(0)

        scrBegin;

        /**
            TODO: parse bisqbot-rockman.xpath, install it as a tracker
                  (create compiler in bisqbot.cc), poll that object
                  in this routine.
        ***/

        static const unsigned framelimit = 56950-615;
        /* Note: loop and local variables need to be "static"
         * in this function because of the co-routine mechanism.
         * RunFrameWith actually returns from this function, and
         * goto's to the same spot after the frame has been run.
         */
        #define NXT { if(BqtDamaged) \
                        { fail_line = __LINE__; goto WalkFail; } \
                      if(framecount >= framelimit) \
                        { fail_line = __LINE__; goto NextFail; } }

        #define RunFramesWhile(condition, input, maxframes, if_fail) \
           ({ static signed long cnt, maxv; \
              for(maxv=(maxframes), cnt=0; (condition); ++cnt) { \
                if(cnt >= maxv) { fail_line = __LINE__; if_fail; } \
                static unsigned in; \
                in = (input); \
                /* This hack adds walking (in the beginning of the clip) \
                 * to the input. The XS comparison checks if Rockman is \
                 * deaccelerating, and pushes Right once to instantly \
                 * decelerate to almost halt. \
                 */ \
                if(walk_done < walk_length) { in |= K_R; } \
                else if(BqtRockXS == 0x0080) { in |= K_R; } \
                ++walk_done; \
                RunFrameWith(in); \
                /* Load updated coordinates */ \
                /* BqtBossHit, BqtBossRemHp, BqtSelPushed, BqtDamaged \
                 * and BqtFiring are updated by emulation core (by tracking \
                 * which ROM addresses are executed. They can't be read from RAM.) \
                 */ \
                BqtRockY = RAM[0x600]; \
                BqtRockXS = RAM[0x4C0]*256+RAM[0x4E0]; \
                NXT; \
              } \
              cnt; /* return value */ \
            })

        static unsigned prefix;
        static unsigned frames_best;
        static SaveState begin_anchor, best;
        static int fail_line;
        begin_anchor.Create();
        frames_best=9999999;
        fprintf(stderr, "Frame counter at start: %u\n", (unsigned)framecount);
        
        for(prefix=37; prefix<=80; ++prefix)
        {
            static unsigned walk_length;
            for(walk_length=0; walk_length<60/*prefix+18*/; ++walk_length)
            {
                __label__ NextFail;
                static std::string current_option;
                { char Buf[512];
                  sprintf(Buf, "idle=%u[walk=%u]",
                    prefix, walk_length);
                  current_option = Buf;
                }

                begin_anchor.Load();
                BqtRockY=255;
                BqtFiring=0;
                BqtBossHit=0;
                BqtDamaged=0;
                
                /* Sit still (or walk), then jump to the height that works,
                 * and shoot, and wait until the boss registered the damage.
                 */
                static unsigned walk_done, jump_done;
                walk_done = 0; jump_done = 0;
                RunFramesWhile(true, 0,   prefix, break);
                RunFramesWhile(BqtRockY > 134, K_A, 26,   goto FailPrefix);
                RunFramesWhile(!BqtFiring, K_B, 3, goto FailPrefix);
                RunFramesWhile(!BqtBossHit, 0, 20, goto FailPrefix);
                /* If we requested more walking than there was action frames
                 * useable, break the loop because the work was already done once. */
                if(walk_done < walk_length) break;
                
                static SaveState action_begin;
                action_begin.Create();
                
                static unsigned undu_delay; // number of loops to skip from the undulation cycle
                static unsigned begin_pausestart; // number of times to use Start in the beginning
                static unsigned end_pausestart;   // number of times to use Start in the end
                static unsigned end_skip;         // number of times to not use Start in the end, after the Starts
                static unsigned pausestart_hp;    // save this value so it can be restored in each loop
                pausestart_hp = BqtBossRemHp;
                for(undu_delay=1; undu_delay-->0; )
                  for(begin_pausestart = 0; begin_pausestart < 4; ++begin_pausestart)
                    for(end_pausestart = 0; end_pausestart < 8-begin_pausestart; ++end_pausestart)
                      for(end_skip=0; end_skip < 4; ++end_skip)
                      {
                          if(end_skip > 0 && end_pausestart == 0) break;
                          
                          __label__ NextFail;
                          static std::string pause_option;
                          { char Buf[512];
                            sprintf(Buf, ",start=[%u,%u,%u,%u]",
                              begin_pausestart, end_pausestart, end_skip, undu_delay);
                            pause_option = current_option + Buf;
                          }

                          BqtBossRemHp = pausestart_hp;
                          action_begin.Load();
                  
                          static SaveState last_action; /* savestate for movie ending frame candidate */
                          
                          /* As long as boss has HP left, run this loop. */
                          /* At the beginning of each loop, the game is running
                           * with the boss just having suffered a hit.
                           */
                          static unsigned hit_counter, undu_counter;
                          undu_counter = 0;
                          for(hit_counter = 0; BqtBossRemHp; ++hit_counter)
                          {
                              if(framecount + (BqtBossRemHp-1)*7 > framelimit)
                                  { fail_line = __LINE__;
                                    if(BqtBossRemHp > end_pausestart+end_skip)
                                        goto FailEndPausePoint;
                                    goto NextFail; }
                              
                              /* Hold Select until the game gets paused
                               * (may take longer than 1 frame due to lag).
                               */
                              BqtSelPushed=0;
                              RunFramesWhile(!BqtSelPushed, K_SE, 3, goto FailPauseMethod);
                              
                              /* Check which method we will use to unpause */
                              static bool UnpauseWithStart;
                              UnpauseWithStart = false;
                              if(hit_counter < begin_pausestart) UnpauseWithStart = true;
                              else if(end_skip && BqtBossRemHp == 1) UnpauseWithStart = false;
                              else if(BqtBossRemHp < end_pausestart + end_skip) UnpauseWithStart = true;
                              
                              if(!UnpauseWithStart)
                              {
                                  /* Unpause with SELECT */
                                  RunFrameWith(0);    NXT;
                                  RunFrameWith(K_SE); NXT;
                                  last_action.Create();

                                  /* Wait until boss registered damage */
                                  BqtBossHit=BqtBossRemHp=0;
                                  if(undu_counter >= undu_delay)
                                  {
                                      /* undulate up&down while pausing (Bisqwit's style) */
                                      static SaveState anc2;
                                      static long k;
                                      for(k=0; !BqtBossHit; ++k)
                                      {
                                          if(k >= 27) goto FailPauseMethod;
                                          anc2.Create();
                                          RunFrameWith(K_R);
                                      }
                                      anc2.Load();
                                      RunFrameWith(K_R | K_A);
                                  }
                                  else
                                  {
                                      /* simple version of above */
                                      RunFramesWhile(!BqtBossHit, K_R, 27, goto FailPauseMethod);
                                  }
                                  ++undu_counter;
                              }
                              else
                              {
                                  /* Unpause with START */
                                  /* Check how many frames of idle are required before
                                   * pressing Start will immediately yield damage.
                                   * Immediateness is used to avoid lag.
                                   */
                                  static unsigned n;
                                  for(n=0; ; )
                                  {
                                      static SaveState anc2;
                                      anc2.Create();

                                      /* Try pressing start. */
                                      BqtBossHit=BqtBossRemHp=0;
                                      RunFrameWith(K_ST); NXT;
                                      if(BqtBossHit) { last_action.Create(); break; }
                                      
                                      /* If too many frames have elapsed, either the shot
                                       * won't hit the boss at all, or it will take too long.
                                       */
                                      if(++n >= 10) { fprintf(stderr, "Missed the boss! hitcounter=%u\n", hit_counter);
                                                      fail_line = __LINE__; goto FailEndPausePoint; }

                                      /* If the boss didn't suffer, undo Start, idle 1 frame, and try again. */
                                      anc2.Load();
                                      RunFrameWith(0); NXT;
                                  }
                              }
                          } /* repeat until the boss's HP is exhausted */
                          
                          last_action.Load(); /* Shorten the movie to minimal length */
                          
                          fprintf(stderr, "With %s, it ends at frame %u!\n",
                              pause_option.c_str(), (unsigned)framecount);
                          if(framecount <= frames_best)
                          {
                              frames_best = framecount;
                              best.Create();
                              FCEUSS_Save("botfront1-middle");
                          }
                          
                          continue;
                      FailEndPausePoint:
                          end_pausestart=100; end_skip=10;
                      FailPauseMethod:
                      NextFail:
                          fprintf(stderr, "%s is not acceptable (fail_line=%d,framecount=%u)\n",
                              pause_option.c_str(), fail_line, (unsigned)framecount);
                      }
                
                continue;
            WalkFail:
                walk_length=999;
            FailPrefix:
            NextFail:
                fprintf(stderr,
                    "%s is not acceptable (fail_line=%d, Y=%d, frno=%u)\n",
                    current_option.c_str(), fail_line, BqtRockY, (unsigned)framecount);
            }
        }
    EndLoop: ;
        fprintf(stderr, "Loading best savestate: %u frames\n", frames_best);
        best.Load();
        
        #undef RunFrameWith
        scrFinishV;
    }
};

struct GenericLuckManipulator
{
    struct Attempt
    {
        std::vector<unsigned char> Seq;
        SaveState anchor;
        unsigned retry_count;
        unsigned retry_max;
        unsigned frame_begin;
        unsigned frame_count;
        unsigned frame_max;
        
        void Init(unsigned Step, unsigned Frameno)
        {
            Seq.resize(Frameno);
            anchor.Create();
            retry_count = 0;
            retry_max   = MAX_REFINES[Step];
            frame_begin = Frameno;
            frame_count = Frameno;
            frame_max   = MAX_LENGTH[Step];
        }
        void Reset()
        {
            Seq.resize(frame_begin);
            frame_count = frame_begin;
            anchor.Load();
            if(frame_count == 0) BisqBotReset();
            ++retry_count;
        }
    };

    void Run()
    {
        scrBegin;
        
        BotFrontMsg("--------------- New run  ------------------ \n");

        #define RunFrameWith(k) do { \
            CurInput = (k); \
            FCEUI_FrameAdvance(); \
            scrReturnV; if(!Active)goto EndLoop; } while(0)
        
        static std::vector<unsigned char> BestSeq;
        static unsigned long long BestScore;
        
        static Attempt Steps[MAX_PARTS];
        static unsigned CurStep;
        
        CurStep=0;
        Steps[0].Init(0, 0);
        Steps[0].Reset();

        BestSeq.clear();
        BestScore = 0;
        
        /* MAIN LOOP */
        for(;;)
        {
            /* All frames handled? */
            if(Steps[CurStep].frame_count >= Steps[CurStep].frame_max)
            {
            TryOver:
                /* Return to the beginning of this step (and possibly,
                 * to the end of the previous step)
                 */
                Steps[CurStep].Reset();

                /* Check retry counter */
                if(Steps[CurStep].retry_count < Steps[CurStep].retry_max)
                {
                    /* One more time */
                    ++Steps[CurStep].retry_count;
                    continue;
                }
                /* All retries exhausted */
                if(CurStep > 0)
                {
                    BotFrontMsg("*** Return step %u->%u\n", CurStep,CurStep-1);
                    /* Back to previous level */
                    --CurStep;
                    goto TryOver;
                }
                /* End loop */
                break;
            }
            
            /* Run frame. Scope the "k" off. */
            if(true)
            {
                GlobalSeqSample = Steps[CurStep].Seq;

                static unsigned char k;
                //fprintf(stderr, "Frame %u\r", Steps[CurStep].frame_count);
                k = BisqBotInput(Steps[CurStep].frame_count++);
                Steps[CurStep].Seq.push_back(k);
                RunFrameWith(k);
            }
            
            /* Let BisqBot analyze how we're doing */
            unsigned long long Score = BisqBotEvaluate(Steps[CurStep].frame_count);
            bool Restart=false;
            switch(BisqBotTick(Steps[CurStep].frame_count, CurStep))
            {
                case BisqBotRestart:
                    CurStep=0; // passthru
                case BisqBotRestartStep:
                {
                    /* Try over */
                    Restart=true;
                    break;
                }
                case BisqBotContinue:
                {
                    /* ok */
                    break;
                }
                case BisqBotNextStep:
                {
                    /* Go to next step */
                    if(CurStep+1 < MAX_PARTS)
                    {
                        BotFrontMsg("*** Advance step %u->%u at frame %u (score %llu)\n",
                            CurStep,CurStep+1, Steps[CurStep].frame_count, Score);
                        DisplaySeq(Steps[CurStep].Seq);
                        BotFrontMsg("\n");

                        Steps[CurStep+1].Init(CurStep+1, Steps[CurStep].frame_count);
                        Steps[CurStep+1].Seq = Steps[CurStep].Seq;
                        ++CurStep;
                    }
                    break;
                }
                case BisqBotAccept:
                {
                    /* Bingo */
                    const char* beep = "";
                    if(Score >= BestScore) beep = "";
                    
                    BotFrontMsg("\n*** %s Bingo in %u frames (score: %llu):\n",
                        beep, Steps[CurStep].frame_count, Score);
                    DisplaySeq(Steps[CurStep].Seq);
                    BotFrontMsg("\n\n");
                    
                    Restart=true;
                    break;
                }
            }
            if(Score > BestScore)
            {
                BestScore = Score;
                BestSeq   = Steps[CurStep].Seq;
            }
            if(Restart) goto TryOver;
        }
        
        /* Lastly, apply the best sequence. Shield with a scope again. */
        if(true)
        {
            Steps[0].Reset();
            
            static unsigned frameno;
            for(frameno = 0; frameno < BestSeq.size(); ++frameno)
            {
                RunFrameWith(BestSeq[frameno]);
            }
        }
        BotFrontEnd();
        
    EndLoop: ;    
        #undef RunFrameWith
        scrFinishV;
    }
};

struct LunarballLaunchManager
{
    void DecideAngleAdjustMethod(unsigned wanted, unsigned char& key, bool& autofire)
    {
        unsigned orig_angle = RAM[0x3C0];
        unsigned plus_dist  = (wanted+256-orig_angle) & 255;
        unsigned minus_dist = (orig_angle+256-wanted) & 255;
        int dir = plus_dist > minus_dist ? -1 : 1;
        unsigned dist = std::min(plus_dist, minus_dist);
        key = dir == 1 ? K_R : K_L;
        unsigned elapse_norepeat = 16 + (dist-1)*1;
        unsigned elapse_repeat = 1 + 2*(dist-1);
        autofire = elapse_repeat <= elapse_norepeat;
    }

    void Run()
    {
        scrBegin;

        #define RunFrameWith(k) do { CurInput = (k); FCEUI_FrameAdvance(); scrReturnV; if(!Active)goto EndLoop; } while(0)
        
        static bool first_round;
        first_round = true;

BeginNewShot:
        static SaveState itbegin, beststate;
        itbegin.Create();
        
        static const unsigned maxframes = 60*60;
        
        static unsigned best_pocketed, best_time, best_model;
        best_pocketed=best_time=best_model=0;
        
#if 0
        if(first_round)
        {
            best_pocketed=2;
            best_time    =645;
            best_model   =67/*angle*/ + 210*256/*velo*/;
        }
#endif
        
        static unsigned angle;
        angle=0;
        
        //if(first_round) angle=0;

        for(; angle<256; ++angle)
        {
            //if(first_round && angle == 47) angle = 51;
            /*
            if(angle >= 0xF8) continue;
            if(angle < 0x28) continue;
            if(angle != 0x6B) continue;
            */
        
            itbegin.Load();

            static unsigned frames_begin;
            static unsigned char angle_key;
            static bool angle_autofire;
            DecideAngleAdjustMethod(angle, angle_key, angle_autofire);
            
            frames_begin = 0;
            if(angle_autofire)
            {
                // autofire
                for(;;) { RunFrameWith(angle_key); ++frames_begin;
                          if(RAM[0x3C0] == angle) break;
                          RunFrameWith(0); ++frames_begin; }
            }
            else
            {
                // hold button until angle is right
                while(RAM[0x3C0] != angle) { RunFrameWith(angle_key); ++frames_begin; }
            }
            
            static std::vector<bool> covered_velocities;
            covered_velocities.clear();
            covered_velocities.resize(256, false);
            for(;;)
            {
                for(unsigned a=0; a<256; a+=3)
                    if(!covered_velocities[a])
                    {
                        //fprintf(stderr, "velocity %u not covered yet...\n", a);
                        goto not_all_velo_covered;
                    }
                //fprintf(stderr, "next angle\n");
                break;
            not_all_velo_covered: ;
                static unsigned velo;
                velo = RAM[0x3A0]; // Check which velocity is in turn
                // If we have already handled it for this angle, wait 1 frame and try again
                if(covered_velocities[velo]) { RunFrameWith(0); ++frames_begin; continue; }
                covered_velocities[velo] = true;

#if 1 /* IGNORE LOW VELOCITIES */
                if(velo < 160) continue;
#endif
                
                static unsigned velobegin_frames;
                static SaveState velobegin_state;
                velobegin_frames = frames_begin;
                velobegin_state.Create();
                
                fprintf(stderr, "\rang[%3d]velo[%3d][%3u]", angle, velo, frames_begin);
                fflush(stderr);
            
                // Count how many balls were pocketed before playing
                static unsigned pocketed_begin;
                pocketed_begin=0;
                for(unsigned a=0; a<16;++a)
                {
                    if(RAM[0x300+a]==0xC0) break;
                    if((RAM[0x300+a] & 15) != 2 && (RAM[0x300+a] & 15) != 1)
                    {
                        //fprintf(stderr, "%u:%02X;", a,RAM[0x300+a]);
                        ++pocketed_begin;
                    }
                }

                // Fire (and wait until it registers the action)
                if(!frames_begin) RunFrameWith(0);
                do { RunFrameWith(K_B); ++frames_begin;
                } while(RAM[0x3D0] == 0 || RAM[0x3D0] == 0xFF);

                static unsigned num_pocketed;
                num_pocketed=0;
                
                static const char* failreason;
                failreason=0;
    
                // Wait until all balls stop moving.
                static unsigned nframes;
                nframes=0;
                while(RAM[0x3D0] != 0 && RAM[0x3D0] != 0xFF)
                {
                    if(RAM[0x3A0] == 0)
                    {
                        // If white ball is not moving, check if all other balls
                        // have been pocketed
                        for(unsigned b=1; b<=6; ++b)
                            if(RAM[0x300+b] >= 4) goto NotAllPocketed;
                        break; // yay!
                    NotAllPocketed:;
                    }
                
                    if(RAM[0x300]%16 == 3) { failreason="white pocketed #1"; goto Fail; }
                    if(++nframes > maxframes) { failreason="timeout"; goto Fail; }
                    //if(frames_begin+nframes >= 300) goto Fail;
                    RunFrameWith(0);
                }
                nframes += frames_begin;
                
                for(unsigned a=0; a<16;++a)
                {
                    if(RAM[0x300+a]==0xC0) break;
                    if((RAM[0x300+a] & 15) != 2
                    && (RAM[0x300+a] & 15) != 1
                    )
                    {
                        ++num_pocketed;
                        if(a == 0) { failreason="white pocketed #2"; goto Fail; }
                    }
                }
                
                /*fprintf(stderr, "frame %u, pocketed %u, begin %u\n",
                    (unsigned)framecount,
                    num_pocketed,
                    pocketed_begin);
                */
                
                if(num_pocketed >= pocketed_begin)
                    num_pocketed -= pocketed_begin; else num_pocketed=0;
                
                if(num_pocketed > best_pocketed
                || (num_pocketed == best_pocketed && nframes < best_time)
                  )
                {
                    best_pocketed = num_pocketed;
                    best_time   = nframes;
                    best_model = angle + velo*256;
                    fprintf(stderr, "Record: Pocketed %u (time %u, angle %u, velo %u)\n",
                        best_pocketed, best_time, angle, velo);
                    beststate.Create();
                    FCEUSS_Save("botfrontL-middle");
                }
                //if(!num_pocketed) failreason="nothing pocketed";
                
                Fail:;
                //if(failreason)
                //    fprintf(stderr, "fail: %s\n", failreason);
                
                velobegin_state.Load();
                frames_begin = velobegin_frames;
            }
        }
        first_round = false;
        
        beststate.Load();
        /*
        After the shot has been shot, delay until as long as pressing K_L
        will affect the aimer. Then proceed with the next shot.
        */
        if(true)
        {
            static SaveState tmp;
            static unsigned was_angle, became_angle1, became_angle2;
            
            fprintf(stderr, "Waiting until next shot can be shot\n");
        NextFrameWaiter:
            tmp.Create();
            was_angle = RAM[0x3C0];
            RunFrameWith(K_R);
            became_angle1 = RAM[0x3C0];
            if(became_angle1 == was_angle) goto NextFrameWaiter;
            tmp.Load();
            RunFrameWith(K_L);
            became_angle2 = RAM[0x3C0];
            if(became_angle2 == was_angle) goto NextFrameWaiter;
            /* If Right and Left produced the same change, reject this action */
            if(became_angle1 == became_angle2) goto NextFrameWaiter;
            
            tmp.Load();
            
            fprintf(stderr, "Beginning new shot\n");
            FCEUSS_Save("botfrontL-newshotbegin");
            goto BeginNewShot;
        }
        
    EndLoop: ;    
        BotFrontEnd();
        beststate.Load();
        FCEUSS_Save("botfrontL-final");
        
        #undef RunFrameWith
        scrFinishV;
    }
};

void BotFrontIter()
{
    scrBegin;
Restart:
    while(!Active)
    {
        BisqBotEvaluate(0);
        BisqBotReset();
        scrReturnV;
    }

    if(!FCEUI_EmulationPaused()) FCEUI_ToggleEmulationPause();

#if 0 /* CV2 first 50 hearts */
#endif

#if 0 /* LUNAR BALL PLAYER */
    static LunarballLaunchManager code;
    while(Active) { code.Run(); scrReturnV; }
    BotFrontEnd();
    FCEUSS_Save("botfrontL-end");
    goto Restart;
#endif

#if 0 /* CUTMAN GLITCH NAVIGATOR v2 */
    static CutmanGlitchNavigatorv2 code;
    while(Active) { code.Run(); scrReturnV; }
    BotFrontEnd();
    FCEUSS_Save("botfront2-end");
    goto Restart;
#endif

#if 0 /* CUTMAN GLITCH NAVIGATOR v1 */
    static CutmanGlitchNavigatorv1 code;
    while(Active) { code.Run(); scrReturnV; }
    BotFrontEnd();
    FCEUSS_Save("botfront-end");
    goto Restart;
#endif

#if 0 /* WILY BOSS FIGHTER */
    static WilyBossFighter code;
    while(Active) { code.Run(); scrReturnV; }
    BotFrontEnd();
    FCEUSS_Save("botfront-end");
    goto Restart;
#endif

#if 1 /* GENERIC CORE */
    static GenericLuckManipulator gencore;
    while(Active) { gencore.Run(); scrReturnV; }
    BotFrontEnd();
    FCEUSS_Save("botfront-end");
    goto Restart;
#endif

    goto Restart;
    
    scrFinishV;
}
