#include "tiletracker.hh"

#if NESVIDEOS_TRACKING

#include <vector>
#include <map>
#include <gd.h>

#include "uncertainpixel.hh"

#include "doubledeque.hh"

typedef unsigned int uint32 ;
typedef unsigned char uint8 ;
typedef unsigned short uint16;

#if PIXEL_METHOD==PIXEL_METHOD_CHANGELOG
unsigned CurrentTimer = 0;
#endif

class TILE_Tracker
{
    /* In this tracker, alpha = visibility. 0=transparent */
    static const unsigned max_x = 256;
    static const unsigned max_y = 256;
    static const unsigned max = max_x * max_y;
    uint16 CurrentScreen[max];
    
    class Layer
    {
        int org_x, org_y;
        
        int xmin,ymin;
        int xmax,ymax;
        bool first;
        
        unsigned count;
        
        int get_min_y() const { return ymin; }
        int get_max_y() const { return ymax; }
        int get_min_x() const { return xmin; }
        int get_max_x() const { return xmax; }
        
        DoubleDeque<UncertainPixel> screens;
        
        const std::vector<uint32> LoadScreen(int ox,int oy, unsigned sx,unsigned sy)
        {
            std::vector<uint32> result(sy*sx, 0);
            unsigned pos=0;
            for(unsigned y=0; y<sy; ++y)
            {
                const DoubleDeque<UncertainPixel>::Row& row = screens.getRow(y+oy);
                for(unsigned x=0; x<sx; ++x)
                    result[pos++] = row.at(x+ox);
            }
            return result;
        }
        void PutScreen(const uint16*const input, int ox,int oy, unsigned sx,unsigned sy)
        {
            unsigned pos = 0;
            for(unsigned y=0; y<sy; ++y)
                for(unsigned x=0; x<sx; ++x)
                {
                    unsigned pix16 = input[pos++];
                    unsigned r = (pix16 >> (11-3)) & 0xF8;
                    unsigned g = (pix16 >> (5-2))  & 0xFC;
                    unsigned b = (pix16 << -(0-3)) & 0xF8;
                    screens.at(y+oy, x+ox).set(r,g,b);
                }
        }
        
        void Save()
        {
            if(first) return;
            
            #if PIXEL_METHOD==PIXEL_METHOD_CHANGELOG
                static bool Saving = false;
                if(!Saving)
                {
                    Saving = true;
                    unsigned SavedTimer = CurrentTimer;
                    for(CurrentTimer=0; CurrentTimer<SavedTimer; ++CurrentTimer)
                    {
                        count = CurrentTimer;
                        Save();
                    }
                    CurrentTimer = SavedTimer;
                    Saving = false;
                }
            #endif
            
            int ymi = get_min_y(), yma = get_max_y();
            int xmi = get_min_x(), xma = get_max_x();

            unsigned wid = xma-xmi+1;
            unsigned hei = yma-ymi+1;
            
            if(wid == 1 || hei == 1) return;
            
            fprintf(stderr, " (%d,%d)-(%d,%d)\n", xmi,ymi, xma,yma);
            
            std::vector<uint32> screen = LoadScreen(xmi,ymi, wid,hei);

            gdImagePtr im = gdImageCreateTrueColor(wid,hei);
            
            for(unsigned p=0, y=0; y<hei; ++y)
                for(unsigned x=0; x<wid; ++x)
                {
                    unsigned pix32 = screen[p++];
                    unsigned pix = pix32/*gdTrueColor
                    (
                        (pix32 >> 24) & 0xFF,
                        (pix32 >> 16) & 0xFF,
                        (pix32 >> 8) & 0xFF
                    )*/;
                    gdImageSetPixel(im, x,y, pix);
                }
            
            char Filename[512];
            sprintf(Filename, "tile-%04u.png", count);
            
            FILE* fp = fopen(Filename, "wb");
            gdImagePng(im, fp);
            fclose(fp);
            gdImageDestroy(im);
        }
        
    public:
        Layer() : count(0)
        {
            Reset();
        }
        ~Layer()
        {
        }
        
        void Reset()
        {
            Save();
            
            fprintf(stderr, " Too different\n");
            screens.clear();
            org_x = 0x40000000;
            org_y = 0x40000000;
            ++count;
            first = true;
        }

        void FitScreen(const uint16* buf, int bestdx, int bestdy, bool suspect)
        {
            static unsigned framecounter=0;
            if(++framecounter == 600) { Save(); framecounter=0; }
            
            if(bestdx != 0 || bestdy != 0)
            {
                fprintf(stderr, " Motion(%d,%d), Origo(%d,%d)\n", bestdx,bestdy, org_x,org_y);
            }

            org_x += bestdx; org_y += bestdy;
            
            if(suspect)
            {
                std::vector<uint32> oldbuf = LoadScreen(org_x,org_y, max_x,max_y);
                unsigned diff = 0;
                for(unsigned a=0; a<oldbuf.size(); ++a)
                {
                    unsigned pix   = oldbuf[a];
                    unsigned pix16 = buf[a];
                    unsigned r = (pix16 >> (11-3)) & 0xF8;
                    unsigned g = (pix16 >> (5-2))  & 0xFC;
                    unsigned b = (pix16 << -(0-3)) & 0xF8;
                    unsigned oldr = ((pix >> 16) & 0xFF);
                    unsigned oldg = ((pix >>  8) & 0xFF);
                    unsigned oldb = ((pix >>  0) & 0xFF);
                    int rdiff = (int)(r-oldr); if(rdiff < 0)rdiff=-rdiff;
                    int gdiff = (int)(g-oldg); if(gdiff < 0)gdiff=-gdiff;
                    int bdiff = (int)(b-oldb); if(bdiff < 0)bdiff=-bdiff;
                    unsigned absdiff = rdiff+gdiff+bdiff;
                    diff += absdiff;
                }
                if(diff > max_x * 256)
                {
                    Reset();
                }
            }
            
            if(first || org_x < xmin) xmin = org_x;
            if(first || org_y < ymin) ymin = org_y;
            int xtmp = org_x+max_x; if(first || xtmp > xmax) xmax=xtmp;
            int ytmp = org_y+max_y; if(first || ytmp > ymax) ymax=ytmp;
            first=false;
            
            PutScreen(buf, org_x,org_y, max_x,max_y);
        }
        
        void SaveNow()
        {
            Save();
        }
        
    } layer;
    
public:
    TILE_Tracker()
    {
    }
    
    void Reset()
    {
        layer.Reset();
    }
    
    void Begin()
    {
        for(unsigned a=0; a<max; ++a) CurrentScreen[a]=0x0000;
    }
    void End(int xdiff, int ydiff, bool suspect)
    {
       // fprintf(stderr, "End frame %u\n", n);
        /* Analyze motion */
        layer.FitScreen(CurrentScreen, xdiff,ydiff, suspect);
    }
    inline void Set(unsigned pos, uint16 pixel)
    {
        if(pos >= max) return;  /* silently ignore bad coordinate */
        //fprintf(stderr, "[%u]", pos);
        CurrentScreen[pos] = pixel;
    }
    void Save()
    {
        layer.SaveNow();
    }
} TILE_Tracker;

extern "C" { /******* PPU variables: **/

/* 0..7. Finetuning the scrolling. */
unsigned VScroll=0;
unsigned HScroll=0;
unsigned NTAWrites=0;

void TILE_TrackFrame(const void*data, unsigned width, unsigned height)
{
    TILE_Tracker.Begin();

    const unsigned short * buf = (const unsigned short *)data;
    unsigned pos=0;
    for(unsigned y=0; y<height; ++y)
        for(unsigned x=0; x<width; ++x, ++pos)
            TILE_Tracker.Set(pos, buf[pos]);

    static unsigned PrevV;
    static unsigned PrevH;
    static bool first=true;
    if(first) { PrevV=VScroll; PrevH=HScroll; first=false; }
    
    fprintf(stderr, "%d,%d\t", HScroll,VScroll);

    int vdiff = (VScroll+240-PrevV)%240; PrevV=VScroll;
    int hdiff = (HScroll+256-PrevH)%256; PrevH=HScroll;
    
    if(hdiff >= 128) hdiff -= 256;
    if(vdiff >= 120) vdiff -= 240;
    
    long absdiff2 = (hdiff*hdiff) + (vdiff*vdiff);
    
    bool suspect = false;
    if(/*absdiff2 > 20*20 || */NTAWrites > 200)
    {
        suspect = true;
    }
    
/*
    if(NTAWrites != 0)
        fprintf(stderr, "%ld NTA writes\n", NTAWrites);
*/
    TILE_Tracker.End(hdiff, vdiff, suspect);
    NTAWrites = 0;

#if PIXEL_METHOD==PIXEL_METHOD_CHANGELOG
    ++CurrentTimer;
#endif

}
void TILE_ForceSave()
{
    TILE_Tracker.Save();
}


}

#else
extern "C" { 
void TILE_ForceSave() { }
}

#endif
