#ifndef bqtRockman1Sound
#define bqtRockman1Sound

#include <vector>
#include <stdarg.h>

#include "midi.hh"
#include "midi.cc"

class Rockman1Sound
{
public:
    void Update()
    {
        IterateMusic();
        IterateSFX();
        NextFrame();

        static unsigned b=0;
        if(++b==60) { b=0; Save("test.mid"); }
    }
    void Command(unsigned char cmd, signed char param)
    {
        switch(cmd)
        {
            case 0xFC: SetSpeedUp((unsigned char)param + 1); break;
            case 0xFD: SetFading(param); break;
            case 0xFE: Reset(); break;
            case 0xFF: StopMusic(); break;
            default: InitSound(cmd);
        }
    }
private:
    class MIDI: public MIDIfile
    {
    public:
        MIDI()
        {
            tempo = 1000000; // 1 second
            deltaticks = 60; // 60 ticks per second
            InitTrack(0);
            InitTrack(1);
            balance = 0;
            for(unsigned c=0; c<16; ++c)
                for(unsigned n=0; n<256; ++n)
                    NotesOn[c][n]=false;
        }
        void Stop(MIDIfile& target, unsigned trackno)
        {
            MIDIfile::tracks.clear();
            
            for(unsigned c=0; c<16; ++c)
                for(unsigned n=0; n<256; ++n)
                    if(NotesOn[c][n])
                    {
                        target.KeyOff(trackno, c, n, 127);
                        NotesOn[c][n] = false;
                    }
        }
        void Load(const std::string& fn)
        {
            std::string ffn = "/mnt/oc/mid/megaman/" + fn;
            fprintf(stderr, "Loading %s...\n", ffn.c_str());
            bool ok = MIDIfile::Load(ffn.c_str());
            fprintf(stderr, "-> %s\n", ok?"ok":"not ok");
            
            for(unsigned a=0; a<MIDIfile::tracks.size(); ++a)
                MIDIfile::tracks[a].ReadDelay();
        
            balance = 0;
        }
        void NextFrame()
        {
            if(MIDIfile::tracks.size() >= 1) AddDelay(0, 1); // track 0, 1 frame
            if(MIDIfile::tracks.size() >= 2) AddDelay(1, 1); // track 1, 1 frame
        }
        void IterateTo(MIDI& target, unsigned trackno)
        {
            long long target_delta = target.deltaticks;
            long long wanted = target.tempo * deltaticks;
            balance += wanted;
            while(balance >= wanted)
            {
                for(unsigned a=0; a<MIDIfile::tracks.size(); ++a)
                {
                    MIDItrack& track = MIDIfile::tracks[a];
                    IterateTrackTo(track, target, trackno);
                }

                long long used = tempo * target_delta;
                balance -= used;
            }
        }
    private:
        bool NotesOn[16][256];
        long long balance;
        
        void IterateTrackTo(MIDItrack& track, MIDI& target, unsigned trackno)
        {
            if(track.GetDelay() > 0) track.DecDelay(1);
            while(track.GetDelay() == 0)
            {
                std::string metadata;
                int metatype = track.ReadMeta(metadata);
                switch(metatype)
                {
                    case -1:
                    {
                        track.PeekNextCmd(metadata);
                        
                        unsigned cmd  = metadata[0]&0xF0;
                        unsigned chan = metadata[0]&0x0F;
                        switch(cmd)
                        {
                            case 0x90: NotesOn[chan][(unsigned char)metadata[1]] = true; break;
                            case 0x80: NotesOn[chan][(unsigned char)metadata[1]] = false; break;
                        }
                        track.ReadEventInto(target, trackno);
                        break;
                    }
                    case 0x2F: // end track
                        break;
                    case 0x51: // tempo
                    {
                        tempo =
                            ((unsigned char)metadata[0])*256*256+
                            ((unsigned char)metadata[1])*256+
                            ((unsigned char)metadata[2]);
                        break;
                    }
                }
                track.ReadDelay();
            }
        }
    } sfx, music, result;
private:
    void InitSound(unsigned char cmd)
    {
        if(cmd <= 0x10) StopMusic();
        else StopSFX();
        std::string midifn = "";
        switch(cmd)
        {
            case 0x00: break;
            case 0x01:
                midifn = "mm1stageselect.mid";
                break;
            case 0x02:
                midifn = "mm1bossselected.mid";
                break;
            case 0x03: break;
            case 0x04:
                midifn = "mm1bossdefeat.mid";
                break;
            case 0x05: break;
            case 0x06: break;
            case 0x07: break;
            case 0x08: break;
            case 0x09:
                midifn = "MM1ELEC4.MID";
                break;
            case 0x0A: break;
            case 0x0B: break;
            case 0x0C: break;
            case 0x0D:
                midifn = "MMBOSS.MID";
                break;
            case 0x0E: break;
            case 0x0F: break;
            case 0x10: break;
            case 0x11: break;
            case 0x12: break;
            case 0x13: // mag beam
                break;
            case 0x14: // shoot P
                break;
            case 0x15: break;
            case 0x16: // megaman hit
                break;
            case 0x17: break;
            case 0x18: break;
            case 0x19: // hit ground
                 break;
            case 0x1A: // acquire magnetbeam
                break;
            case 0x1B: // enemy kill
                break;
            case 0x1C: // game start, boss launch
                break;
            case 0x1D: break;
            case 0x1E: break;
            case 0x1F: // menu select
                break;
            case 0x20: // transform
                break;
            case 0x21: // score shuffle
                break;
            case 0x22: // pause sound
                break;
            case 0x23: break;
            case 0x24: break;
            case 0x25: break;
            case 0x26: break;
            case 0x27: break;
            case 0x28: break;
            case 0x29: break;
            case 0x2A: break;
            case 0x2B: break;
            case 0x2C: // pneuma block
                break;
            case 0x2D: break;
            case 0x2E: break;
            case 0x2F: break;
            case 0x30: break;
            case 0x31: break;
            case 0x32: break;
        }
        if(midifn.empty())
        {
            fprintf(stderr, "Sound 0x%02X not implemented\n", cmd);
            return;
        }
        if(cmd <= 0x10) music.Load(midifn); else sfx.Load(midifn);
    }
    
    void SetSpeedUp(unsigned/* value*/)
    {
        // unused in Rockman
    }
    void SetFading(int/* param*/)
    {
        // unused in Rockman
    }
    void Reset()
    {
        StopMusic();
        StopSFX();
    }
    void StopMusic()
    {
        music.Stop(result, 0);
    }
    void StopSFX()
    {
        sfx.Stop(result, 1);
    }
    void IterateMusic()
    {
        music.IterateTo(result, 0);
    }
    void IterateSFX()
    {
        sfx.IterateTo(result, 1);
    }
    void NextFrame()
    {
       // music.NextFrame();
       // sfx.NextFrame();
        result.NextFrame();
    }
    void Save(const char* fn)
    {
        MIDIfile m = result;
        m.Done();
        m.Save("test.mid");
    }
};

#endif

