/* FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <zlib.h>

#include "types.h"
#include "video.h"
#include "fceu.h"
#include "general.h"
#include "memory.h"
#include "crc32.h"
#include "state.h"
#include "movie.h"
#include "palette.h"
#include "nsf.h"
#include "input.h"
#include "vsuni.h"

uint8 *XBuf=NULL;
uint8 *XBackBuf=NULL;
static uint8 *xbsave=NULL;

void FCEU_KillVirtualVideo(void)
{
 if(xbsave)
 {
  free(xbsave);
  xbsave=0;
 }
 if(XBackBuf)
 {
  free(XBackBuf);
  XBackBuf=0;
 }
}

int FCEU_InitVirtualVideo(void)
{
 if(!XBuf)		/* Some driver code may allocate XBuf externally. */
			/* 256 bytes per scanline, * 240 scanline maximum, +8 for alignment,
			*/
 if(!(XBuf= (uint8*) (FCEU_malloc(256 * 256 + 8))) ||
    !(XBackBuf= (uint8*) (FCEU_malloc(256 * 256 + 8))))
  return 0;
 xbsave=XBuf;

 if(sizeof(XBuf)==4)
 {
  unsigned m;
  m=(unsigned)XBuf;
  m=(4-m)&3;
  XBuf+=m;
 } 
 memset(XBuf,128,256*256); //*240);
 memset(XBackBuf,128,256*256);
 return 1;
}

int howlong;
static char errmsg[16384*4];

#include "drawing.h"
#ifdef FRAMESKIP

//#define SHOWFPS
void ShowFPS(void);
void FCEU_PutImageDummy(void)
{
 #ifdef SHOWFPS
 ShowFPS();
 #endif
 if(FCEUGameInfo->type!=GIT_NSF)
 {
  FCEU_DrawNTSCControlBars(XBuf);
  FCEU_DrawSaveStates(XBuf);
  FCEU_DrawMovies(XBuf);
 }
 if(howlong) howlong--; /* DrawMessage() */
}
#endif

static int dosnapsave=0;
void FCEUI_SaveSnapshot(void)
{
 dosnapsave=1;
}
 
static void FCEU_DrawRecordingStatus(uint8* XBuf)
{
	int n=0;
#ifdef WIN32
	if(FCEUD_ShowStatusIcon())
#else
    if(0)
#endif
	{
		if(FCEUI_EmulationPaused())
			n=3;
		else if(FCEUI_IsMovieActive()>0)
			n=2;
		else if(FCEUI_IsMovieActive()<0)
			n=1;
	}

	if(n)
		FCEU_DrawRecordingStatusN(XBuf, n);
}

static void ReallySnap(void)
{
 int x=SaveSnapshot();
 if(!x)
  FCEU_DispMessage("Error saving screen snapshot.");
 else
  FCEU_DispMessage("Screen snapshot %d saved.",x-1);
}

static void RockmanBeams()
{
    extern unsigned char RefrRAM[0x800];

    unsigned a;
    int scrollbegin = RefrRAM[0x1B]*256 + RefrRAM[0x1A];
    //for(a=0; a<256*240; ++a) XBuf[a] = (XBuf[a] & 15) | 192;
    if(RefrRAM[0x1E] != 0) return; // vscroll
    for(a=0; a<16; ++a)
    {
        unsigned power  = RefrRAM[0x5B0+a];
        int      xmid   = RefrRAM[0x5D0+a]*256 + RefrRAM[0x5C0+a];
        int      ypos   = RefrRAM[0x5F0+a];
        unsigned length = RefrRAM[0x5A0+a];
        unsigned units = 1+length/4;
        if(!power) continue;
        
        int screenbegin = (xmid - scrollbegin) - units*8;
        int screenend   = (xmid - scrollbegin) + units*8;
        
        if(screenbegin < 0) screenbegin = 0;
        if(RefrRAM[0x1A] != 0 && screenend   > 255) screenend = 255;
        if(RefrRAM[0x1A] != 0 && screenbegin > 255) continue;
        if(screenend   < 0) continue;
        
        if(RefrRAM[0x1E] != 0) ypos -= RefrRAM[0x1E]-256;
        
        while(screenbegin <= screenend)
        {
            unsigned x=screenbegin&255;
            int y, ycount=0;
            for(y=ypos-2; y<ypos+4; ++y,++ycount)
            {
                int color=12;
                if(screenbegin > 255)
                {
                    continue;//if((x|y)&1)continue;
                }
                if(y < 0 || y > 239) continue;
                switch(ycount)
                {
                    case 0: color=13; break;
                    case 5: color=13; break;
                    default:
                    {
                        int offs = y+screenbegin+RefrRAM[0x1A];
                        int hi=0x2C, lo=0x11;
                        //if(power==156 && (RefrRAM[0x23]&1)) hi=0x3C;
                        //  -RefrRAM[0x23]/8; // frame counter for visfx
                        color = (offs&2) ? hi:lo;
                    }
                }
                //if(((x^y)&1) != (RefrRAM[0x23]&1)) continue;
                XBuf[x + y*256] = color | 128;
            }
            ++screenbegin;
        }
    }
}

extern void DrawTextTrans(uint8 *dest, uint32 width, uint8 *textmsg, uint8 fgcolor);

static void PutBlade(double xpos, double ypos, unsigned Index, unsigned phase, unsigned op)
{
    char Buf[64]; 
    unsigned x,y;

    unsigned color = 0x3C;          // NOW
    if(phase > Index) color = 0x2D; // old
    if(phase < Index) color = 0x19; // future
    
    xpos+=1; ypos+=1;
    
    for(y=2; y<=13; ++y)
        for(x=2; x<=13; ++x)
        {
            unsigned yp = ypos*16 + y;
            unsigned xp = xpos*16 + x;
            
            if(x==2 || y==2 || x==13 || y==13)
            {
                unsigned darker = color;
                if(color >= 0x10) darker -= 0x10;
                XBuf[xp + yp*256] = darker | 0x80;
            }
        }
    
    sprintf(Buf, "%X", Index);
    DrawTextTrans(
        &XBuf[ (unsigned)(xpos*16+1) + 256*(unsigned)(ypos*16+4) ],
        256,
        Buf,
        color | 0x80);
}

static void Wily2Blades()
{
    unsigned bosshp = RAM[0x6C1];
    unsigned phase = (28-bosshp) / 2;
    unsigned a;
    
    static int Enable = 0;
    if(!Enable)
    {
        double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        double RockX  =
            RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
        if(RockX >= 10108 && RockY <= 16) Enable=1;
        if(!Enable) return;
    }
    
    /* 2: Tmnhetkiset
     * 1: Tulevat
     * 0: Muut 
     */
    
    #define pb(x,y, Index) \
        if( (a==2 && Index==phase) \
        ||  (a==1 && Index>phase) \
        ||  (a==0 && Index<phase)) \
            PutBlade(x,y, Index, phase,a);
    
    for(a=0; a<3; ++a)
    {
        pb(3.0,  7, 13);
        pb(4.0,  7, 13);
        pb(4.2,  6, 13);

        pb(6.8,  9, 12);
        pb(5.2,  9, 12);
        pb(7.5,  9, 12);

        pb(11.0, 7, 11);
        pb(9.0, 6.5, 11);

        pb(5.0,  7, 10);
        pb(3.0,  6, 10);
        pb(6.0,  6, 10);

        pb(6.5,  8, 9);
        pb(5.2,  8, 9);
        pb(7.8,  8, 9);

        pb(2.0,  7, 8);
        pb(4.0,  6, 8);
        pb(4.0, 6.5, 8);

        pb(9.0,  7, 7);
        pb(10.0, 6, 7);
        pb(7.0,  6, 7);

        // 6: too far
        
        pb(6.5,  7, 5);
        pb(7.1,  7, 5);
        pb(5.7,  7, 5);
        
        pb(10.0, 7.2, 4);
        pb(8.0,   7, 4);

        pb(4.01,  7, 3);
        pb(6.0,  7, 3);
        
        pb(6.5,  9, 2);
        pb(5.2, 10, 2);
        pb(7.8, 10, 2);

        pb(6.0,  8, 1);
        pb(7.0,  8, 1);

        // 0: too far
    }
}

#include "drivers/pc/tiletracker.hh"

void FCEU_PutImage(void)
{
extern int BotFrontDisableVideo;
if(BotFrontDisableVideo)
{
  memcpy(XBackBuf, XBuf, 256*256);
  return;
}



	#ifdef SHOWFPS
	ShowFPS();
	#endif
#if 1
    RockmanBeams();
    //Wily2Blades();
#endif
        if(FCEUGameInfo->type==GIT_NSF)
        {
         DrawNSF(XBuf);
         /* Save snapshot after NSF screen is drawn.  Why would we want to
            do it before?
         */
         if(dosnapsave)
         {
          ReallySnap();
          dosnapsave=0;
         }
        } 
        else
        {   
         /* Update AVI before overlay stuff is written */
         /* Save backbuffer before overlay stuff is written. */
         if(!FCEUI_EmulationPaused())
          memcpy(XBackBuf, XBuf, 256*256);
         /* Save snapshot before overlay stuff is written. */
         if(dosnapsave)
         {
          ReallySnap();
          dosnapsave=0;
         }
         if(FCEUGameInfo->type==GIT_VSUNI)
          FCEU_VSUniDraw(XBuf);
	 FCEU_DrawSaveStates(XBuf);
	 FCEU_DrawMovies(XBuf);
         FCEU_DrawNTSCControlBars(XBuf);
		 FCEU_DrawRecordingStatus(XBuf);
        }
#ifndef NESVIDEOS_TRACKING
        DrawMessage();
#endif
        FCEU_DrawInput(XBuf);
}

void FCEU_DispMessage(char *format, ...)
{
 va_list ap;

 va_start(ap,format);
 vsprintf(errmsg,format,ap);
 va_end(ap);

 va_start(ap,format);
 vfprintf(stderr, format, ap);
 fprintf(stderr, "\n");
 va_end(ap);

 howlong=40;
 //howlong=0; //DISABLE MESSAGES

 extern int movcounter;
 movcounter=0;
}

void FCEU_ResetMessages(void)
{
 howlong=0;
}


static int WritePNGChunk(FILE *fp, uint32 size, char *type, uint8 *data)
{
 uint32 crc;

 uint8 tempo[4];

 tempo[0]=size>>24;
 tempo[1]=size>>16;
 tempo[2]=size>>8;
 tempo[3]=size;

 if(fwrite(tempo,4,1,fp)!=1)
  return 0;
 if(fwrite(type,4,1,fp)!=1)
  return 0;

 if(size)
  if(fwrite(data,1,size,fp)!=size)
   return 0;

 crc=CalcCRC32(0,(uint8 *)type,4);
 if(size)
  crc=CalcCRC32(crc,data,size);

 tempo[0]=crc>>24;
 tempo[1]=crc>>16;
 tempo[2]=crc>>8;
 tempo[3]=crc;

 if(fwrite(tempo,4,1,fp)!=1)
  return 0;
 return 1;
}

int SaveSnapshot(void)
{
 static unsigned int lastu=0;

 char *fn=0;
 int totallines=FSettings.LastSLine-FSettings.FirstSLine+1;
 int x,u,y;
 FILE *pp=NULL;
 uint8 *compmem=NULL;
 uLongf compmemsize=totallines*263+12;

 if(!(compmem=(uint8 *)FCEU_malloc(compmemsize)))
  return 0;

 for(u=lastu;u<99999;u++)
 {
  pp=FCEUD_UTF8fopen((fn=FCEU_MakeFName(FCEUMKF_SNAP,u,"png")),"rb");
  if(pp==NULL) break;
  fclose(pp);
 }

 lastu=u;

 if(!(pp=FCEUD_UTF8fopen(fn,"wb")))
 {
  free(fn);
  return 0;
 }
 free(fn);
 {
  static uint8 header[8]={137,80,78,71,13,10,26,10};
  if(fwrite(header,8,1,pp)!=1)
   goto PNGerr;
 }

 {
  uint8 chunko[13];

  chunko[0]=chunko[1]=chunko[3]=0;
  chunko[2]=0x1;			// Width of 256

  chunko[4]=chunko[5]=chunko[6]=0;
  chunko[7]=totallines;			// Height

  chunko[8]=8;				// bit depth
  chunko[9]=3;				// Color type; indexed 8-bit
  chunko[10]=0;				// compression: deflate
  chunko[11]=0;				// Basic adapative filter set(though none are used).
  chunko[12]=0;				// No interlace.

  if(!WritePNGChunk(pp,13,"IHDR",chunko))
   goto PNGerr;
 }

 {
  uint8 pdata[256*3];
  for(x=0;x<256;x++)
   FCEUD_GetPalette(x,pdata+x*3,pdata+x*3+1,pdata+x*3+2);
  if(!WritePNGChunk(pp,256*3,"PLTE",pdata))
   goto PNGerr;
 }

 {
  uint8 *tmp=XBuf+FSettings.FirstSLine*256;
  uint8 *dest,*mal,*mork;

  if(!(mal=mork=dest=(uint8 *)malloc((totallines<<8)+totallines)))
   goto PNGerr;
 //   mork=dest=XBuf;

  for(y=0;y<totallines;y++)
  {
   *dest=0;			// No filter.
   dest++;
   for(x=256;x;x--,tmp++,dest++)
    *dest=*tmp; 	
  }

  if(compress(compmem,&compmemsize,mork,(totallines<<8)+totallines)!=Z_OK)
  {
   if(mal) free(mal);
   goto PNGerr;
  }
  if(mal) free(mal);
  if(!WritePNGChunk(pp,compmemsize,"IDAT",compmem))
   goto PNGerr;
 }
 if(!WritePNGChunk(pp,0,"IEND",0))
  goto PNGerr;

 free(compmem);
 fclose(pp);

 return u+1;


 PNGerr:
 if(compmem)
  free(compmem);
 if(pp)
  fclose(pp);
 return(0);
}
#ifdef SHOWFPS
uint64 FCEUD_GetTime(void);
uint64 FCEUD_GetTimeFreq(void);

static uint64 boop[60];
static int boopcount = 0;

void ShowFPS(void)
{ 
 uint64 da = FCEUD_GetTime() - boop[boopcount];
 char fpsmsg[16];
 int booplimit = PAL?50:60;
 boop[boopcount] = FCEUD_GetTime();

 sprintf(fpsmsg, "%8.1f",(double)booplimit / ((double)da / FCEUD_GetTimeFreq()));
 DrawTextTrans(XBuf + (256-8-8*8) + (FSettings.FirstSLine+4)*256,256,fpsmsg,4);
 // It's not averaging FPS over exactly 1 second, but it's close enough.
 boopcount = (boopcount + 1) % booplimit;
}
#endif
